//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     parse.cxx
// Purpose:    Syntax Analysis
// Last Change:12.09.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2000-2001 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "out.h"
#include "check.h"
#include "opt.h"
#include "atom.h"
#include "cnst.h"
#include "var.h"
#include "scope.h"
#include "unsafe.h"
#include "term.h"
#include "tlist.h"
#include "lit.h"
#include "op.h"
#include "fml.h"
#include "flist.h"
#include "conv.h"
#include "ans.h"

#include "parse.h"

//=============================================================================
// Class Variables:
//=============================================================================

//-----------------------------------------------------------------------------
// Lex: The input token stream for this file (interface to scanner).
//-----------------------------------------------------------------------------

lex_c*	parse_c::Lex = 0;

//-----------------------------------------------------------------------------
// Tok: The type of the current token (i.e. Lex->tok_type()).
//-----------------------------------------------------------------------------

tok_t	parse_c::Tok = TOK_EOF;

//-----------------------------------------------------------------------------
// NumErrs: The number of syntax errors discovered so far.
//-----------------------------------------------------------------------------

int	parse_c::NumErrs = 0;

//-----------------------------------------------------------------------------
// OpStack: Operator Stack.
//-----------------------------------------------------------------------------

op_c	*parse_c::OpStack[PARSE_STACKSIZE];

//-----------------------------------------------------------------------------
// OpStackPtr: Stackpointer for operator stack (one above top element).
//-----------------------------------------------------------------------------

op_c	**parse_c::OpStackPtr = 0;

//-----------------------------------------------------------------------------
// OpStackEnd: Pointer to one above last possible element of operator stack.
//-----------------------------------------------------------------------------

op_c	**parse_c::OpStackEnd = 0;

//-----------------------------------------------------------------------------
// FmlStack: Operand (Formula) Stack.
//-----------------------------------------------------------------------------

fml_c	*parse_c::FmlStack[PARSE_STACKSIZE];

//-----------------------------------------------------------------------------
// FmlStackPtr: Stackpointer for operand stack (one above top element).
//-----------------------------------------------------------------------------

fml_c	**parse_c::FmlStackPtr = 0;

//-----------------------------------------------------------------------------
// FmlStackEnd: Pointer to one above last possible element of operand stack.
//-----------------------------------------------------------------------------

fml_c	**parse_c::FmlStackEnd = 0;

//-----------------------------------------------------------------------------
// LogOpWithinNot: AND or OR were used inside default negation.
//-----------------------------------------------------------------------------

#ifdef VER_SEM_STABLE
bool_t	parse_c::LogOpWithinNot = BOOL_FALSE;
#endif

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Parse a file:
//-----------------------------------------------------------------------------

flist_c *parse_c::parse(in_c* input_file)
{
	fml_c *fml;
	flist_c *input_formulas;
	flist_c *last;
	flist_c *new_last;

	// Check Parameter:
	CHECK(input_file->is_open(), "parse_c::parse: Input stream not open");

	// Initialize data members:
	Lex	= new lex_c(input_file);
	NumErrs = 0;
	Tok	= Lex->tok_type();
	OpStackPtr  = OpStack;
	OpStackEnd  = OpStack + PARSE_STACKSIZE;
	FmlStackPtr = FmlStack;
	FmlStackEnd = FmlStack + PARSE_STACKSIZE;
#ifdef VER_SEM_STABLE
	LogOpWithinNot = BOOL_FALSE;
#endif

	// Parse formulas:
	input_formulas = 0;
	last = 0;
	while(Tok != TOK_EOF && NumErrs <= PARSE_MAX_ERRS) {
		if(Tok == TOK_CMD) {
			if(parse_cmd() == BOOL_FALSE) {
				// Error: Skip remaining tokens in formula:
				while(Tok != TOK_EOF && Tok != TOK_FULLSTOP
					&& Tok != TOK_NOSPACE)
					next();
				// Now skip the fullstop:
				if(Tok == TOK_FULLSTOP || Tok == TOK_NOSPACE)
				next();
				}
		}
		else {
			fml = parse_fml_or_query();
			if(fml == 0) {
				// Error: Skip remaining tokens in formula:
				while(Tok != TOK_EOF && Tok != TOK_FULLSTOP
					&& Tok != TOK_NOSPACE)
					next();
				// Now skip the fullstop:
				if(Tok == TOK_FULLSTOP || Tok == TOK_NOSPACE)
					next();
			}
			else if(input_formulas == 0) {
				// Construct formula list:
				input_formulas = new flist_c(fml);
				last = input_formulas;
			}
			else {
				// Append formula to existing formula list:
				new_last = new flist_c(fml);
				last->append(new_last);
				last = new_last;
			}
		}
	}

	// If there were errors, print summary:
	if(NumErrs > PARSE_MAX_ERRS) {
		out_c::err_end(-1);
		return(0);
	}
	else if(NumErrs > 0) {
		out_c::err_end(NumErrs);
		return(0);
	}
	// Else check for empty input:
	else if(input_formulas == 0) {
		out_c::err_input_empty();
		return(0);
	}
	// Else print formula list:
	else {
		return(input_formulas);
	}
}

//=============================================================================
// Auxillary Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// error: Print error message for current token.
//-----------------------------------------------------------------------------

void parse_c::error(out_syn_t msg)
{
	// Begin error message list if this is the first error:
	if(NumErrs == 0)
		out_c::err_begin();

	// Count number of errors:
	NumErrs++;

	// The real reason might be a lexical error:
	// (The scanner prints no error message, it returns an error token.)
	switch(Tok) {
		case TOK_ILLEGAL_CHAR:
			msg =OUT_ILLEGAL_CHAR;
			break;
		case TOK_LESS:
			msg = OUT_BAD_LESS;
			break;
		case TOK_GREATER:
			msg = OUT_BAD_GREATER;
			break;
		case TOK_MINUS:
			msg = OUT_BAD_MINUS;
			break;
		case TOK_MISSINGQ:
			msg = OUT_MISSING_QUOTE;
			break;
		case TOK_EMPTYA:
			msg = OUT_EMPTY_ATOM;
			break;
		case TOK_NOSPACE:
			msg = OUT_MISSING_SPACE;
			break;
		case TOK_INT_OVERFLOW:
			msg = OUT_INT_OVERFLOW;
			break;
		case TOK_EMPTY_CMD:
			msg = OUT_EMPTY_CMD;
			break;
		case TOK_RESERVED_ATOM:
			msg = OUT_RESERVED_PRED;
			break;
		default:
			// The given message is correct.
			break;
	}

	// Print syntax error message:
	out_c::err_syntax(Lex->filename(), Lex->line_no(), msg);
	if(Lex->is_v())
		out_c::err_note_v();
	out_c::err_newline();
	Lex->show();
}

//-----------------------------------------------------------------------------
// error_head: Print error message for invalid '_' or 'not' in the head.
//-----------------------------------------------------------------------------

void parse_c::error_head(fml_t fml)
{
	// Begin error message list if this is the first error:
	if(NumErrs == 0)
		out_c::err_begin();

	// Count number of errors:
	NumErrs++;

	// Print error message:
	if(fml->is_lit())
		out_c::err_anon_head(Lex->filename(), Lex->line_no());
	else {
		CHECK(fml->get_op()->kind() == OP_DEFNEG,
			"parse_c::error_head_lit: Unexpected formula");
		out_c::err_neg_head(Lex->filename(), Lex->line_no());
	}
	fml->print();
	out_c::err_newline();
	Lex->show();
}

//-----------------------------------------------------------------------------
// error_var: Print error message for unsafe variables:
//-----------------------------------------------------------------------------

void parse_c::error_var(void)
{
	// Begin error message list if this is the first error:
	if(NumErrs == 0)
		out_c::err_begin();

	// Count number of errors:
	NumErrs++;

	// Print error message:
	out_c::err_unbound_var(Lex->filename(), Lex->line_no());
	unsafe_c::print();
	out_c::err_newline();
	Lex->show();
}

//-----------------------------------------------------------------------------
// parse_pred: Parse a predicate.
//-----------------------------------------------------------------------------

atom_c *parse_c::parse_pred(void)
{
	atom_c *pred;

	// An atom is expected here:
	if(Tok != TOK_ATOM) {
		error(OUT_PRED_EXPECTED);
		return(0);
	}

	// Ok, so get pointer to atom:
	pred = atom_c::find(Lex->tok_ptr(), Lex->tok_end());

	// Check for reserved atoms:
	if(pred == atom_c::answer()) {
		error(OUT_RESERVED_PRED);
		return(0);
	}

	// Ok, token is successfully processed. Return predicate:
	next();
	return(pred);
}

//-----------------------------------------------------------------------------
// parse_term: Parse a term.
//-----------------------------------------------------------------------------

term_c *parse_c::parse_term(void)
{
	atom_c *atom;
	var_c *var;
	term_c *term;
	int n;

	// Check next token:
	if(Tok!=TOK_ATOM && Tok!=TOK_INT && Tok!=TOK_VAR && Tok!=TOK_ANONVAR) {
		error(OUT_TERM_EXPECTED);
		return(0);
	}

	// Treat different kinds of terms:
	switch(Tok) {
		case TOK_ATOM:
			atom = atom_c::find(Lex->tok_ptr(), Lex->tok_end());
			term = new term_c(new cnst_c(atom));
			break;
		case TOK_VAR:
			var = scope_c::find(Lex->tok_ptr(), Lex->tok_end());
			if(var == 0) {
				error(OUT_TOO_MANY_VARS);
				return(0);
			}
			term = new term_c(var);
			break;
		case TOK_INT:
			n = Lex->tok_int();
			term = new term_c(new cnst_c(n));
			break;
		case TOK_ANONVAR:
			if(within_defneg()) {
				error(OUT_ANON_WITHIN_NOT);
				return(0);
			}
			term = new term_c();
			break;
		default:
			term = 0;
			CHECK_IMPOSSIBLE(
				"parse_c::parse_term: Unexpected token");
	}

	// Token processed:
	next();

	// Return term:
	return(term);
}

//-----------------------------------------------------------------------------
// parse_args: Parse arguments of a literal (list of terms).
//-----------------------------------------------------------------------------

tlist_c *parse_c::parse_args(void)
{
	term_c  *term;
	tlist_c *args;
	tlist_c *last;
	tlist_c *new_last;
	int num_args;

	// Get first term:
	term = parse_term();
	if(term == 0)
		return(0);

	// Check whether argument terms are allowed at all:
	num_args = 1;
	if(num_args > VER_MAX_ARGS) {
		error(OUT_TOO_MANY_ARGS);
		return(0);
	}

	// Create term list object;
	args = new tlist_c(term);
	last = args;

	// Check continuation:
	if(Tok != TOK_COMMA && Tok != TOK_RPAREN) {
		error(OUT_COMMA_OR_RPAREN_EXPECTED);
		return(0);
	}

	// Read rest of terms:
	while(Tok == TOK_COMMA) {
		// Skip Comma:
		next();

		// Read term:
		term = parse_term();
		if(term == 0)
			return(0);

		// Check limit on number of argument terms:
		num_args++;
		if(num_args > VER_MAX_ARGS) {
			error(OUT_TOO_MANY_ARGS);
			return(0);
		}

		// Add to argument list:
		new_last = new tlist_c(term);
		last->append(new_last);
		last = new_last;

		// Check continuation:
		if(Tok != TOK_COMMA && Tok != TOK_RPAREN) {
			error(OUT_COMMA_OR_RPAREN_EXPECTED);
			return(0);
		}
	}

	// Return argument list:
	return(args);
}

//-----------------------------------------------------------------------------
// parse_lit: Parse a literal.
//-----------------------------------------------------------------------------

lit_c *parse_c::parse_lit(bool_t parenthesized)
{
	atom_c *pred;
	tlist_c *args;
	lit_c *lit;

	// Get the predicate:
	pred = parse_pred();
	if(pred == 0)
		return(0);

	// Check next token:
	if(parenthesized && Tok!=TOK_LPAREN && Tok!=TOK_BIN_OP
		&& Tok!=TOK_COMMA && Tok!=TOK_RPAREN) {
		error(OUT_ARGLIST_BINOP_OR_RPAREN_EXPECTED);
		return(0);
	}
	if(!parenthesized && Tok!=TOK_LPAREN && Tok!=TOK_BIN_OP
		&& Tok!=TOK_COMMA && Tok!=TOK_FULLSTOP) {
		error(OUT_ARGLIST_BINOP_OR_FULLSTOP_EXPECTED);
		return(0);
	}

	// Get argument list:
	if(Tok == TOK_LPAREN) {
		// Skip left parenthesis:
		next();

		// Read argument list:
		args = parse_args();
		if(args == 0)
			return(0);

		// Now the next token must be a right parenthesis:
		if(Tok != TOK_RPAREN) {
			error(OUT_RPAREN_EXPECTED);
			return(0);
		}

		// Skip right parenthesis:
		next();
	}
	else
		args = 0;

	// Create literal and return it:
	lit = new lit_c(pred, args);
	return(lit);
}

//-----------------------------------------------------------------------------
// within_defneg: We currently reading the argument of a not operator.
//-----------------------------------------------------------------------------

bool_t parse_c::within_defneg(void)
{
	op_t *p;

	// We only have to check whether there is currently a default negation
	// operator on the stack.
	// All operators which were already read from the input, but do
	// not yet have their operands complete, must be on the operator stack.
	// Therefore, as long as we are reading the operand of a not,
	// that operator is still on the stack.
	// Note that it might be below a NIL operator in case parentheses
	// were used.

	// Search the stack for a default negation operator:
	for(p = OpStack; p < OpStackPtr; p++) {
		if(*p != 0 && (*p)->kind() == OP_DEFNEG)
			return(BOOL_TRUE);
	}

	// Not found:
	return(BOOL_FALSE);
}

//-----------------------------------------------------------------------------
// pop_ops: Pop operators from the stack until specified priority.
//-----------------------------------------------------------------------------

void parse_c::pop_ops(int prio)
// operators up to and including priority prio are popped.
{
	op_c *top_op;
	fml_c *fml, *fml_left, *fml_right;

	// Check consistency of stacks:
	CHECK(OpStackPtr != OpStack, "parse_c::pop_ops: Operator Stack empty");
	CHECK(OpStack <= OpStackPtr && OpStackPtr <= OpStackEnd,
		"parse_c::pop_ops: Operator stack pointer out of range");
	CHECK(FmlStack <= FmlStackPtr && FmlStackPtr <= FmlStackEnd,
		"parse_c::pop_ops: Operand stack pointer out of range");

	// Check parameter:
	CHECK(prio >= 0, "parse_c::pop_ops: Priority too small");
	CHECK(prio <= OP_MAX_PRIO, "parse_c::pop_ops: Priority too large");

	// Pop all operators which bind stronger than the new one.
	// NIL at the bottom of the stack will certainly not be popped:
	top_op = *(OpStackPtr - 1);

	while(top_op != 0 && top_op->prio() <= prio) {
		// Pop operator and operands, construct formula, push it:
		CHECK(FmlStackPtr > FmlStack,
			"parse_c::pop_ops: No right operand");
		fml_right = *--FmlStackPtr;
		CHECK(FmlStackPtr > FmlStack,
			"parse_c::pop_ops: No left operand");
		fml_left  = *--FmlStackPtr;
		if(top_op->is_binary())
			fml = new fml_c(top_op, fml_left, fml_right);
		else
			fml = new fml_c(top_op, fml_right);
		--OpStackPtr;
		top_op = *(OpStackPtr - 1);
		*FmlStackPtr++ = fml;
	}
}

//-----------------------------------------------------------------------------
// push_op: Push an operator. Pop previous ones if necessary.
//-----------------------------------------------------------------------------

bool_t parse_c::push_op(op_c *op)
{
	op_t top_op;

	// Check consistency of stacks:
	CHECK(OpStackPtr != OpStack, "parse_c::push_op: Operator Stack empty");
	CHECK(OpStack <= OpStackPtr && OpStackPtr <= OpStackEnd,
		"parse_c::push_op: Operator stack pointer out of range");
	CHECK(FmlStack <= FmlStackPtr && FmlStackPtr <= FmlStackEnd,
		"parse_c::push_op: Operand stack pointer out of range");

	// Check parameter:
	CHECK(op != 0, "parse_c::push_op: NIL operator");
	CHECK(op->valid(), "parse_c::push_op: Illegal operator");

	// Pop operators on top of the stack if necessary:
	// E.g. an OR (prio=3) can lie below and AND (prio=2),
	// but not vice versa
	// (AND has higher precedence and must be evaluated first).
	switch(op->assoc()) {
		case ASSOC_LEFT:
			// In this case, operators of lower and the same
			// priority (i.e. higher or same precedence)
			// must first be popped:
			pop_ops(op->prio());
			break;
		case ASSOC_RIGHT:
			// In this case, only operators of lower priority
			// (i.e. strictly higher binding strength)
			// must be popped.
			pop_ops(op->prio() - 1);
			break;
		case ASSOC_NONE:
			// In this case, we pop operators of lower priority
			// (strictly higher binding strength)
			// and then check that there is no operator of the
			// same priority on top of the stack:
			pop_ops(op->prio() - 1);
			CHECK(OpStackPtr > OpStack,
			      "parse_c::push_op: NIL at stack bottom missing");
			top_op = *(OpStackPtr-1);
			if(top_op != 0 && top_op->prio() == op->prio()) {
				error(OUT_OPERATOR_NOT_ASSOCIATIVE);
				return(BOOL_FALSE);
			}
			break;
	}

	// Check restrictions for operators within default negation:
	// Note that this test must be done after giving the default negation
	// on the stack a chance to pop.
	switch(op->kind()) {
		case OP_DEFNEG:
			if(within_defneg()) {
				error(OUT_NOT_CANNOT_BE_NESTED);
				return(BOOL_FALSE);
			}
			break;
		case OP_NOT:
		case OP_IF:
		case OP_THEN:
		case OP_EQUIV:
			if(within_defneg()) {
#ifdef VER_SEM_STABLE
				if(opt_c::is_set(OPT_SEM_STABLE))
					error(OUT_NO_OPS_INSIDE_NOT);
				else
#endif
					error(OUT_OP_ILLEGAL_WITHIN_NOT);
				return(BOOL_FALSE);
			}
			break;
		case OP_AND:
		case OP_OR:
			// These operators are legal inside default negation
			// for the static semantics. For the stable semantics,
			// they are currently not supported.
#ifdef VER_SEM_STABLE
			if(within_defneg()) {
				LogOpWithinNot = BOOL_TRUE;
				if(opt_c::is_set(OPT_SEM_STABLE)) {
					error(OUT_NO_OPS_INSIDE_NOT);
					return(BOOL_FALSE);
				}
			}
#endif // VER_SEM_STABLE
			break;
	}


	// If the operator is monadic, push a NIL left operand:
	if(!op->is_binary()) {
		// Check for stack overflow:
		if(FmlStackPtr == FmlStackEnd) {
			error(OUT_STACK_OVERFLOW);
			return(BOOL_FALSE);
		}
		// Now push it:
		*FmlStackPtr++ = 0;
	}

	// Now push the new operator:
	if(OpStackPtr == OpStackEnd) {
		error(OUT_STACK_OVERFLOW);
		return(BOOL_FALSE);
	}
	*OpStackPtr++ = op;
	return(BOOL_TRUE);
}

//-----------------------------------------------------------------------------
// push_fml: Push an operand (i.e. a formula).
//-----------------------------------------------------------------------------

bool_t parse_c::push_fml(fml_c *fml)
{
	// Check consistency of stack:
	CHECK(FmlStack <= FmlStackPtr && FmlStackPtr <= FmlStackEnd,
		"parse_c::push_fml: Operand stack pointer out of range");

	// Check parameter:
	CHECK(fml != 0, "parse_c::push_fml: formula is NIL");
	CHECK(fml->valid(), "parse_c::push_fml: bad formula");

	// Check for stack overflow:
	if(FmlStackPtr == FmlStackEnd) {
		error(OUT_STACK_OVERFLOW);
		return(BOOL_FALSE);
	}

	// Now push it:
	*FmlStackPtr++ = fml;
	return(BOOL_TRUE);
}

//-----------------------------------------------------------------------------
// pop_fml: Return the final formula.
//-----------------------------------------------------------------------------

fml_c *parse_c::pop_fml(void)
{
	// Check consistency of stack:
	CHECK(FmlStackPtr!=FmlStack, "parse_c::pop_fml: Operand Stack empty");
	CHECK(FmlStack <= FmlStackPtr && FmlStackPtr <= FmlStackEnd,
		"parse_c::pop_fml: Operand stack pointer out of range");

	// The stack cannot be empty:
	CHECK(FmlStackPtr!=FmlStack, "parse_c::pop_fml: No formula left");

	// Return the formula left on the stack:
	return(*--FmlStackPtr);
}

//-----------------------------------------------------------------------------
// parse_fml: Parse a formula (auxillary function).
//-----------------------------------------------------------------------------

fml_c *parse_c::parse_fml(bool_t parenthesized)
{
	bool_t	formula_expected;
	lit_c	*lit;
	fml_c	*fml;

	// At the beginning, a formula is needed,
	// not a binary operator, and not the end.
	// formula_expected is true if the operator stack has as many
	// binary operators as the operand stack has formulas.
	// It is false if there is one formula more than binary operators.
	formula_expected = BOOL_TRUE;

	// Set up new stack bottom:
	*OpStackPtr++ = 0;

	// Read tokens:
	while(Tok != TOK_FULLSTOP && Tok != TOK_RPAREN) {

		// Handle Monadic operators:
		if(Tok == TOK_MON_OP) {
			// A mondadic operator starts a formula:
			if(!formula_expected) {
				error(parenthesized ?
					OUT_BINOP_OR_RPAREN_EXPECTED :
					OUT_BINOP_OR_FULLSTOP_EXPECTED);
				return(0);
			}

			// Put it on operator stack etc.:
			if(!push_op(Lex->tok_op()))
				return(0);

			// Skip monadic operator:
			next();
		}

		// Handle atomic formulas (literals):
		else if(Tok == TOK_ATOM) {
			// Check mode:
			if(!formula_expected) {
				error(parenthesized ?
					OUT_BINOP_OR_RPAREN_EXPECTED :
					OUT_BINOP_OR_FULLSTOP_EXPECTED);
				return(0);
			}

			// Read the formula:
			lit = parse_lit(parenthesized);
			if(lit == 0)
				return(0);
			fml = new fml_c(lit);
			if(!push_fml(fml))
				return(0);

			// Now we expect a binary operator or the end:
			formula_expected = BOOL_FALSE;
		}

		// Handle parenthesized subformulas:
		else if(Tok == TOK_LPAREN) {
			// Check mode:
			if(!formula_expected) {
				error(parenthesized ?
					OUT_BINOP_OR_RPAREN_EXPECTED :
					OUT_BINOP_OR_FULLSTOP_EXPECTED);
				return(0);
			}

			// Skip the left parenthesis:
			next();

			// Read the formula:
			fml = parse_fml(BOOL_TRUE);
			if(fml == 0)
				return(0);
			if(!push_fml(fml))
				return(0);

			// The next token must be a right parenthesis:
			if(Tok != TOK_RPAREN) {
				error(OUT_RPAREN_EXPECTED);
				return(0);
			}
			next();

			// Now we expect a binary operator or the end:
			formula_expected = BOOL_FALSE;
		}

		// Handle binary operators:
		else if(Tok==TOK_BIN_OP || Tok == TOK_COMMA) {
			// Check mode:
			if(formula_expected) {
				error(OUT_FORMULA_EXPECTED);
				return(0);
			}

			// Put it on operator stack etc.:
			if(!push_op(Lex->tok_op()))
				return(0);

			// Skip operator:
			next();

			// Now we expect a formula:
			formula_expected = BOOL_TRUE;
		}

		// Handle errors:
		else {
			if(formula_expected) {
				error(OUT_FORMULA_EXPECTED);
				return(0);
			}
			else {
				error(parenthesized ?
					OUT_BINOP_OR_RPAREN_EXPECTED :
					OUT_BINOP_OR_FULLSTOP_EXPECTED);
				return(0);
			}
		}
	}

	// At the end, we must have a formula:
	if(formula_expected) {
		error(OUT_FORMULA_EXPECTED);
		return(0);
	}

	// Process the operators left on the operator stack:
	pop_ops(OP_MAX_PRIO);

	// Pop NIL at the bottom of the stack:
	--OpStackPtr;
	CHECK(*OpStackPtr == 0,
		"parse_c::parse_fml: Operator stack out of order");

	// Fetch result formula from operand stack and return it:
	fml = pop_fml();
	return(fml);
}

//-----------------------------------------------------------------------------
// parse_query: Parse a query, translate into formula about $answer.
//-----------------------------------------------------------------------------

fml_c *parse_c::parse_query(void)
{
	fml_t	query, head_fml, implication;
	var_t	var;
	term_t	arg;
	tlist_t	arg_list, next_list_elem, last_list_elem;
	lit_t	head;

	// This function should only be called when the current token is '?':
	CHECK(Tok==TOK_QUERY,
		"parse_c::parse_query: Current token is not '?'");

	// Print an error is the query is already defined:
	if(ans_c::is_query_set()) {
		error(OUT_ONLY_ONE_QUERY_PERMITTED);
		return(0);
	}

	// Skip question mark:
	next();

	// Check start of query:
	if(Tok != TOK_MON_OP && Tok != TOK_ATOM && Tok != TOK_LPAREN
		&& Tok != TOK_VAR && Tok != TOK_COLON) {
		error(OUT_FORMULA_OR_VARLIST_EXPECTED);
		return(0);
	}

	// Start list of answer variables:
	ans_c::begin_answer_vars();

	// Check whether answer variables are explicitly given:
	if(Tok == TOK_VAR || Tok == TOK_COLON) {
		// Construct the argument list for the head literal:
		arg_list = 0;
		last_list_elem = 0;
		while(Tok == TOK_VAR) {
			// Add variable to argument list:
			var = scope_c::find(Lex->tok_ptr(), Lex->tok_end());
			if(var == 0) {
				error(OUT_TOO_MANY_VARS);
				return(0);
			}
			if(ans_c::add_answer_var(var) == BOOL_FALSE) {
				error(OUT_TOO_MANY_VARS);
				return(0);
			}
			arg = new term_c(var);
			next_list_elem = new tlist_c(arg);
			if(arg_list == 0)
				arg_list = next_list_elem;
			else
				last_list_elem->append(next_list_elem);
			last_list_elem = next_list_elem;

			// Skip variable:
			next();

			// Skip comma, check token:
			if(Tok != TOK_COMMA && Tok != TOK_COLON) {
				error(OUT_COMMA_OR_COLON_EXPECTED);
				return(0);
			}
			if(Tok == TOK_COMMA) {
				next();
				if(Tok != TOK_VAR) {
					error(OUT_VAR_EXPECTED);
					return(0);
				}
			}
			CHECK(Tok == TOK_VAR || Tok == TOK_COLON,
				"parse_c::parse_query: Tok must be var or :");
		}

		// Skip colon:
		CHECK(Tok == TOK_COLON, "parse_c::parse_query: Tok not colon");
		next();

		// Read the query formula:
		query = parse_fml(BOOL_FALSE);
		if(query == 0)
			return(0);
	}

	// Answer variables are implicitly given:
	else {
		// Read the query:
		query = parse_fml(BOOL_FALSE);
		if(query == 0)
			return(0);

		// Construct the argument list for the head literal:
		var = scope_c::first();
		if(var == 0)
			arg_list = 0;
		else {
			if(ans_c::add_answer_var(var) == BOOL_FALSE) {
				error(OUT_TOO_MANY_VARS);
				return(0);
			}
			arg = new term_c(var);
			arg_list = new tlist_c(arg);
			last_list_elem = arg_list;
			for(var = scope_c::next(); var; var = scope_c::next()){
				if(ans_c::add_answer_var(var) == BOOL_FALSE) {
					error(OUT_TOO_MANY_VARS);
					return(0);
				}
				arg = new term_c(var);
				next_list_elem = new tlist_c(arg);
				last_list_elem->append(next_list_elem);
				last_list_elem = next_list_elem;
			}
		}
	}

	// Ok, query is defined:
	ans_c::end_answer_vars();

	// Construct and return the formula $answer(...) <- query:
	head = new lit_c(atom_c::answer(), arg_list);
	head_fml = new fml_c(head);
	implication = new fml_c(op_c::op_if(), head_fml, query);
	return(implication);
}

//-----------------------------------------------------------------------------
// parse_fml: Parse a formula or query, store as clauses.
//-----------------------------------------------------------------------------

fml_c *parse_c::parse_fml_or_query(void)
{
	fml_c   *fml;

	// Open new scope for variables:
	scope_c::reset();

	// Reset Stacks:
	OpStackPtr = OpStack;
	FmlStackPtr = FmlStack;

	// Call auxillary function:
	if(Tok == TOK_QUERY)
		fml = parse_query();
	else
		fml = parse_fml(BOOL_FALSE);
	if(fml == 0)
		return(0);

	// Check that anonymous variables are only used in rule bodies:
	if(!check_head(fml, 1))
		return(0);

	// Convert into clauses:
	unsafe_c::reset();
	conv_c::conv(fml);
	if(conv_c::overflow()) {
		error(OUT_OVERFLOW_DURING_CONVERSION);
		return(0);
	}
	if(conv_c::too_many_head_lits()) {
		error(OUT_TOO_MANY_HEAD_LITS);
		return(0);
	}
	if(conv_c::too_many_body_lits()) {
		error(OUT_TOO_MANY_BODY_LITS);
		return(0);
	}
	if(conv_c::too_many_neg_lits()) {
		error(OUT_TOO_MANY_NEG_LITS);
		return(0);
	}
	if(unsafe_c::is_empty() == BOOL_FALSE) {
		error_var();
		return(0);
	}

	// The next token must be a fullstop:
	if(Tok != TOK_FULLSTOP) {
		error(OUT_FULLSTOP_EXPECTED);
		return(0);
	}
	next();

	// Return formula:
	return(fml);
}

//-----------------------------------------------------------------------------
// check_head: Check whether '_' or 'not' are used in positive context.
//-----------------------------------------------------------------------------

bool_t parse_c::check_head(fml_t fml, int sign)
// sign== -1 means negative context (anonymous variables allowed).
// sign== +1 means positive context (anonymous variables forbidden).
// sign==  0 means both contexts (forbidden, even after negation).
{
	// Check parameters:
	CHECK(fml != 0, "parse_c::check_head: formula is NIL");
	CHECK(fml->valid(), "parse_c::check_head: formula is invalid");
	CHECK(sign == 1 || sign == -1 || sign == 0,
		"parse_c::check_head: invalid sign");

	// End of recursion if this formula is atomic (a literal):
	if(fml->is_lit()) {
		if((sign==1 || sign==0) && fml->get_lit()->contains_anon()) {
			error_head(fml);
			return(BOOL_FALSE);
		}
		else
			return(BOOL_TRUE);
	}

	// Recursive calls for composed formulas:
	else {
		switch(fml->get_op()->kind()) {
			case OP_NOT:
				return(check_head(fml->get_right(), -sign));
			case OP_AND:
				if(!check_head(fml->get_left(), sign))
					return(BOOL_FALSE);
				return(check_head(fml->get_right(), sign));
			case OP_OR:
				if(!check_head(fml->get_left(), sign))
					return(BOOL_FALSE);
				return(check_head(fml->get_right(), sign));
			case OP_IF:
				if(!check_head(fml->get_left(), sign))
					return(BOOL_FALSE);
				return(check_head(fml->get_right(), -sign));
			case OP_THEN:
				if(!check_head(fml->get_left(), -sign))
					return(BOOL_FALSE);
				return(check_head(fml->get_right(), sign));
			case OP_EQUIV:
				if(!check_head(fml->get_left(), 0))
					return(BOOL_FALSE);
				return(check_head(fml->get_right(), 0));
			case OP_DEFNEG:
				if(sign == 1 || sign == 0) {
					error_head(fml);
					return(BOOL_FALSE);
				}
				return(BOOL_TRUE);
			default:
				CHECK_IMPOSSIBLE(
				  "parse_c::check_head: Unknown operator");
				return(BOOL_TRUE);
		}
	}
}

//-----------------------------------------------------------------------------
// parse_cmd: Parse and process a command.
//-----------------------------------------------------------------------------

bool_t parse_c::parse_cmd(void)
{
	str_t cmd_ptr, cmd_end;
	bool_t ok;

	// This should only be called if the current token is TOK_CMD:
	CHECK(Tok==TOK_CMD, "parse_c::parse_cmd: Tok is not TOK_CMD");

	cmd_ptr = Lex->tok_ptr() + 1;
	cmd_end = Lex->tok_end();
	CHECK(cmd_ptr < cmd_end, "parse_c::parse_cmd: Invalid token string");
	if(cmd_end == cmd_ptr+1 && opt_c::set_opt(*cmd_ptr)) {
		// Ok, a legal option. 
#ifdef VER_SEM_STABLE
		// But check whether logical operators inside not were used
		// and the stable model semantics is selected: That is illegal.
		if(*cmd_ptr == 'a' && LogOpWithinNot) {
			error(OUT_OPS_INSIDE_NOT_WERE_USED);
			return(BOOL_FALSE);
		}
#endif
		next();
	}
	else if(cmd_end == cmd_ptr+1 && *cmd_ptr == 'v') {
		// This means options i,t,f,r,s,m:
		ok = opt_c::set_opt('i');
		CHECK(ok, "parse_c::parse_cmd: Option i is invalid");
		ok = opt_c::set_opt('t');
		CHECK(ok, "parse_c::parse_cmd: Option t is invalid");
		ok = opt_c::set_opt('h');
		CHECK(ok, "parse_c::parse_cmd: Option h is invalid");
		ok = opt_c::set_opt('f');
		CHECK(ok, "parse_c::parse_cmd: Option f is invalid");
		ok = opt_c::set_opt('r');
		CHECK(ok, "parse_c::parse_cmd: Option r is invalid");
		ok = opt_c::set_opt('s');
		CHECK(ok, "parse_c::parse_cmd: Option s is invalid");
		ok = opt_c::set_opt('m');
		CHECK(ok, "parse_c::parse_cmd: Option m is invalid");

		next();
	}
	else {
		error(OUT_UNKNOWN_CMD);
		return(BOOL_FALSE);
	}

	// Skip fullstop after option:
	if(Tok != TOK_FULLSTOP) {
		error(OUT_FULLSTOP_EXPECTED);
		return(BOOL_FALSE);
	}
	next();
	return(BOOL_TRUE);
}

